from pypsql_api.config.types import Session
from pypsql_api.wire.actions_types import Names
from pypsql_api.context import Context
from pypsql_api.wire.back import SSLNo, AuthenticationCleartextPassword, ReadyForQuery, AuthenticationOk, \
    EmptyQueryResponse, DataFrameRowDescription, DataFrameDataRows, CommandComplete
from pypsql_api.wire.front import SSLRequest, StartupMessage, Message, PasswordMessage


def read_ssl_request(context: Context):
    ssl_request = SSLRequest.read(context.input)

    return context.update_mem('ssl_request', ssl_request), Names.WRITE_SSL_NO


def write_ssl_resp_no(context: Context):
    SSLNo().write(context.output)

    return context, Names.READ_STARTUP_MESSAGE


def read_startup_message(context: Context):
    startup_front = StartupMessage.read(context.input)

    print(f"Got startup  {startup_front}")

    session = Session(
        user=startup_front.user, database=startup_front.database, password=''
    )

    context.session = session
    return context.update_mem('session', session), Names.WRITE_PLAIN_TEXT_PASSWORD_REQUEST


def write_plain_text_password_request(context: Context):
    AuthenticationCleartextPassword().write(context.output)
    context.output.flush()

    return context, Names.READ_PLAIN_TEXT_PASSWORD


def read_plain_text_password_request(context: Context):
    m, t = Message.read(context.input)

    if not (m and t):
        return context, None

    if not isinstance(m, PasswordMessage):
        raise Exception(f"UnExpected message {m}")

    session = context.session
    session.password = m.password

    auth_ok, msg = context.auth_handler.handle(session=session)

    if auth_ok:
        return context, Names.WRITE_AUTH_OK
    else:
        raise Exception(f"Password not correct {msg}")


def write_ready_for_query(context: Context):
    ReadyForQuery().write(context.output)
    context.output.flush()

    return context, Names.RECEIVE_COMMAND


def write_auth_ok(context: Context):
    AuthenticationOk().write(context.output)
    context.output.flush()

    return context, Names.READY_FOR_QUERY


def read_receive_command(context: Context):
    m, t = Message.read(context.input)

    print(f">>read_receive_command msg {m}, {t}")
    if not (m and t):
        return context, None

    return context.update_mem('message', m), m.process_name


def read_receive_extended_command(context: Context):
    m, t = Message.read(context.input)

    print(f">>read_receive_extended_command msg {m}, {t}")
    if m is None and t is None:
        print(f"Not m,t  {m}, {t}")
        return context, None

    if m and m.process_name in {Names.EXECUTE, Names.BIND, Names.SYNC, Names.DESCRIBE}:
        return context.update_mem('message', m), m.process_name

    print(f"Expected an extended protocol query message but got {m}, {t}")
    return context, Names.ERROR


def write_empty_response(context: Context):
    EmptyQueryResponse().write(context.output)

    return context, Names.READY_FOR_QUERY


def write_data_frame_response(context: Context):
    df = context.mem['data']
    if df is None:
        raise Exception("We expect a data frame instance here")

    DataFrameRowDescription(df=df).write(context.output)
    rows = DataFrameDataRows(df=df, offset=0, max_rows=1000000).write(context.output)
    CommandComplete(tag=f"SELECT {rows}").write(context.output)

    context.output.flush()

    return context, Names.READY_FOR_QUERY


def read_parse_command(context: Context):
    m, t = Message.read(context.input)



from dataclasses import dataclass
from typing import Callable, Any

import numpy as np
import pandas

from pypsql_api.wire.bytes import WritingIO


@dataclass
class DataType:
    oid: str
    typlen: int

    typsend: Callable[[WritingIO, Any], bytes]


def send_as_str(v: Any, encoding='UTF-8') -> bytes:
    return str(v).encode(encoding)


def boolsend(v: Any, encoding="UTF-8") -> bytes:
    return b't' if v else b'f'


def date_out(v, encoding="UTF-8") -> bytes:
    return v.isoformat().encode(encoding)


INT_MAP = {

    np.char: DataType(oid='25', typlen=1, typsend=send_as_str),
    np.int8: DataType(oid='25', typlen=1, typsend=send_as_str),

    np.int16: DataType(oid='21', typlen=2, typsend=send_as_str),

    np.int32: DataType(oid='23', typlen=4, typsend=send_as_str),

    np.int64: DataType(oid='20', typlen=8, typsend=send_as_str),

}

TYPE_MAP = {
    np.bool: DataType(oid='16', typlen=1, typsend=boolsend),
    np.bool_: DataType(oid='16', typlen=1, typsend=boolsend),

    # char

    np.char: INT_MAP[np.char],

    np.int8: INT_MAP[np.int8],
    np.int16: INT_MAP[np.int16],

    np.int32: INT_MAP[np.int32],

    np.int: INT_MAP[np.int64],
    np.int64: INT_MAP[np.int64],

    np.uint8: INT_MAP[np.int8],
    np.uint16: INT_MAP[np.int16],
    np.uint: INT_MAP[np.int64],

    np.uint32: INT_MAP[np.int32],
    np.uint64: INT_MAP[np.int64],

    np.ubyte: INT_MAP[np.char],
    np.byte: INT_MAP[np.char],

    np.float16: DataType(oid='700', typlen=4, typsend=send_as_str),

    np.float32: DataType(oid='700', typlen=4, typsend=send_as_str),

    np.float64: DataType(oid='701', typlen=8, typsend=send_as_str),

    np.float128: DataType(oid='1700', typlen=-1, typsend=send_as_str),

    # str
    np.object: DataType(oid='25', typlen=-1, typsend=send_as_str),
    np.object_: DataType(oid='25', typlen=-1, typsend=send_as_str),

    np.datetime64: DataType(oid='1114', typlen=8, typsend=send_as_str),

    pandas._libs.tslibs.timestamps.Timestamp: DataType(oid='25', typlen=-1, typsend=send_as_str)
}



# see: https://docs.python.org/3/library/socketserver.html
import socketserver
from abc import abstractmethod, ABC
from decimal import Decimal
from typing import Tuple, Optional, Union, Any

from pandas import DataFrame

from pypsql_api.config.types import Session
from pypsql_api.ext.handlers import AuthHandler, QueryHandler, ExtendedQueryHandler
from pypsql_api.wire.actions_types import Names
from pypsql_api.context import Context
from pypsql_api.wire.back import ErrorResponse
from pypsql_api.wire.bytes import ReadingIO, WritingIO
from pypsql_api.wire.extended.types import Portal, PreparedStatement, Param
from pypsql_api.wire.front import BindMessage, ParseMessage
from pypsql_api.wire.protocol import run_states


def state_runner(existing_state, message_processor):
    state = message_processor.next(existing_state)
    message_processor.is_allowed(state)

    return state


class PsqlHandler(ABC, socketserver.StreamRequestHandler, socketserver.ThreadingMixIn):

    @property
    @abstractmethod
    def auth_handler_cls(self):
        raise Exception("Not implemented")

    @property
    @abstractmethod
    def query_handler_cls(self):
        raise Exception("Not implemented")

    def handle(self):

        input = ReadingIO(self.rfile)
        output = WritingIO(self.wfile)

        try:
            context = Context(
                input=input,
                output=output,
                mem={},

                auth_handler=self.auth_handler_cls(),
                query_handler=self.query_handler_cls(),
                extended_query_handler=self.extended_query_handler_cls(),
                session=None,
                portals={},
                prepared_statements={}

            )

            state_name = Names.READ_SSL_REQUEST

            context, state_name = run_states(context, state_name)


        except Exception as e:
            print(e)
            ErrorResponse.severe(str(e)).write(output)

            raise e

        finally:
            output.close()

        print("End")


class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass


def run(auth_handler_cls, query_handler_cls, extended_query_handler_cls, handler=None, host="", port=55432):
    auth_1 = auth_handler_cls
    query_1 = query_handler_cls
    query_2 = extended_query_handler_cls

    if not handler:
        class Handler(PsqlHandler):
            auth_handler_cls = auth_1
            query_handler_cls = query_1
            extended_query_handler_cls = query_2

        handler = Handler

    with ThreadedTCPServer((host, port), handler) as server:
        # Activate the server; this will keep running until you
        # interrupt the program with Ctrl-C
        print(f"Running postgresql server on {host}:{port}")
        server.serve_forever()


def main():
    class TestAuth(AuthHandler):

        def handle(self, session: Session) -> Tuple[bool, Optional[str]]:
            if session.password == "test":
                return True, None

            return False, None

    class TestQueryHandler(QueryHandler):

        def handle(self, session: Session, sql) -> Tuple[Optional[Union[Any, Optional[DataFrame]]], Optional[Any]]:
            import pandas as pd
            import numpy as np
            from decimal import Decimal
            print(f"Handling query {sql}")
            return pd.DataFrame({'id': [1.232323, 2.1233, np.nan], 'name': [1, 2, 3],
                                 'p': [Decimal(1212323213123), Decimal(121312312321321323121321312321312323213),
                                       Decimal(123123213213213213213)]}), None

    import pandas as pd
    import numpy as np
    data_frame = pd.DataFrame({'id': [1.232323, 2.1233, np.nan], 'name': [1, 2, 3],
                               'p': [Decimal(1212323213123), Decimal(121312312321321323121321312321312323213),
                                     Decimal(123123213213213213213)]})

    class TestExtendedQueryHandler(ExtendedQueryHandler):

        def parse(self, session: Session, msg: ParseMessage) -> Tuple[
            Optional[Union[Any, Optional[PreparedStatement]]], Optional[Any]]:
            return PreparedStatement(
                name="test",
                sql=msg.query,
                parameters=[Param(oid_type=1, index=0, type=None), Param(oid_type=1, index=1, type=None)]
            ), None

        def bind(self, session: Session, prepared_statement: PreparedStatement, bind: BindMessage) -> Tuple[
            Optional[Union[Any, Optional[Portal]]], Optional[Any]]:
            return Portal(
                name=bind.portal_name,
                prepared_statement=prepared_statement,
                execution=None,
                row_description=data_frame,
            ), None

        def execute(self, session: Session, portal: Portal, max_rows: int) -> Tuple[
            Optional[Union[Any, Optional[DataFrame]]], Optional[Any]]:

            return data_frame, None

    run(auth_handler_cls=TestAuth, query_handler_cls=TestQueryHandler,
        extended_query_handler_cls=TestExtendedQueryHandler)


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Dict, Any

from pypsql_api.config.types import Session
from pypsql_api.ext.handlers import AuthHandler, QueryHandler, ExtendedQueryHandler
from pypsql_api.wire.bytes import ReadingIO, WritingIO
from pypsql_api.wire.extended.types import PreparedStatement, Portal


@dataclass
class Context:
    input: ReadingIO
    output: WritingIO

    session: Session

    auth_handler: AuthHandler
    query_handler: QueryHandler
    extended_query_handler: ExtendedQueryHandler

    mem: Dict[str, Any]  # keep memory between tasks

    # The extended query protocol
    prepared_statements: Dict[str, PreparedStatement]
    portals: Dict[str, Portal]

    def update_mem(self, k, v):
        return Context(input=self.input, output=self.output,
                       session=self.session,
                       auth_handler=self.auth_handler,
                       query_handler=self.query_handler,
                       extended_query_handler=self.extended_query_handler,
                       mem={**self.mem, **{k: v}},
                       prepared_statements=self.prepared_statements,
                       portals=self.portals)



"""

Portal

Param:
  oid_type
  index
  type
  is_void?

PreparedStatement
  parameters

Session:
 parepared-statements = {
   'UNNAMED': ...
   'Name': ...
   }

 portals = {
  'UNNAMED': ...
  'Name': ...
 }

Named prepared statements can also be created and accessed at the SQL command level, using PREPARE and EXECUTE.

Destroy any unamed prepared statements

F -> Parse
       query-string
       name optional
        data type
        parameter placeholders

      parameter types acn be 0 or
      shorter than the $1 placeholders

      Note we need to see type void id and ignore

B -> ParseComplete
      create prepared statement
   or ErrorResponse


 DECLARE CURSOR and FETCH

F --> Bind (destroy unnamed portal if any)
     name ==> must exist in the session
     return format
     len(prepared_statement.params) must eq len(bind.params)

     void params must have Null

B --> BindComplete
        create portal

    or ErrorResopnse

F --> Execute
     name ==> portal must exist in the session
     result row count, or zero for all

     No rowdescription, only RowData

     If more rows than result row count
     return PortalSuspend
     and save the result and index to the Portal
     a next execute command should carry on from where
     the portal left off

     CommandComplete on completion
     allowed values:
      CommandComplete, EmptyQueryRespones,
      ErrorResponse, PortalSuspended

F --> Sync
     Reads and closes all portals

Note on any back error
   Do:
     Send ErrorResponse and read till Sync found
     then send ReadyForQuery
"""
import traceback

from pandas import DataFrame

from pypsql_api.wire.actions_types import Names
from pypsql_api.context import Context
from pypsql_api.wire.back import ParseComplete, BindComplete, ErrorResponse, EmptyQueryResponse, DataFrameDataRows, \
    PortalSuspended, CommandComplete, DataFrameRowDescription
from pypsql_api.wire.extended.types import PreparedStatement, Portal, Execution
from pypsql_api.wire.front import ParseMessage, BindMessage, ExecuteMessage, Sync, Message, Describe


def process_parse_command(context: Context):
    """
     Process the query
     then write  ParseComplete, and  send to bind
     Expects a query_parser
    :param context:
    :return:
    """
    msg = context.mem.get('message', None)

    if not isinstance(msg, ParseMessage):
        raise Exception(f"ParseMessage expected but got: {msg}")

    if msg.query:
        msg.query = msg.query.replace('OPERATOR(pg_catalog.~)', '=') \
            .replace('~', '=')

        handler = context.extended_query_handler
        resp, error_msg = handler.parse(session=context.session, msg=msg)

        if isinstance(resp, PreparedStatement):
            context.prepared_statements[msg.name] = resp
            ParseComplete().write(context.output)

            return context, Names.RECEIVE_EXTENDED_QUERY_COMMAND
        else:
            print(f"Error: {resp} {error_msg}")
            return context, Names.ERROR
    else:
        return context, Names.ERROR


def process_bind_command(context: Context):
    msg = context.mem.get('message', None)

    if not isinstance(msg, BindMessage):
        if isinstance(msg, Sync):
            return context, Names.SYNC

        raise Exception(f"Bind expected but got: {msg}")

    handler = context.extended_query_handler
    try:
        prepared_statement = context.prepared_statements[msg.prepared_statement_name]

        resp, _ = handler.bind(session=context.session,
                               prepared_statement=prepared_statement,
                               bind=msg)

        print(f"From BIND got handler response: {resp}")

        if isinstance(resp, Portal):
            context.portals[msg.portal_name] = resp
            print(f"From BIND wrote_portals: {context.portals}")

            BindComplete().write(context.output)

            return context, Names.RECEIVE_EXTENDED_QUERY_COMMAND
    except KeyError as e:
        print(f"No prepared statement was found for {msg.prepared_statement_name} in {context.prepared_statements}")
        ErrorResponse.query(str(e)).write(context.output)
        return context, Names.ERROR
    except Exception as e:
        print(f"Error doing bind process {e}")
        ErrorResponse.query(str(e)).write(context.output)
        return context, Names.RECEIVE_EXTENDED_QUERY_COMMAND

    return context, Names.ERROR


def process_execute_command(context: Context):
    msg = context.mem.get('message', None)

    print(f"Process {msg} in process_execute_command")

    if not isinstance(msg, ExecuteMessage):
        if isinstance(msg, Sync):
            return context, Names.SYNC

        raise Exception(f"ExecuteMessage expected but got: {msg}")

    handler = context.extended_query_handler
    try:
        portal = context.portals[msg.portal_name]

        resp, _ = handler.execute(session=context.session,
                                  portal=portal,
                                  max_rows=msg.max_rows)

        if isinstance(resp, DataFrame):
            execution = portal.execution
            if not execution:
                execution = Execution(
                    max_rows=msg.max_rows if msg.max_rows else 10000000,
                    offset=0,
                    df=resp
                )
                portal.execution = execution

            context = context.update_mem("execution_portal", portal)

            return context, Names.WRITE_EXTENDED_DATA_FRAME

        return context, Names.WRITE_EXTENDED_EMPTY_RESPONSE
    except KeyError:
        error_msg = f"No portal name was found for {msg.portal_name}"
        print(error_msg)
        ErrorResponse.query(error_msg).write(context.output)
        return context, Names.SYNC
    except Exception as e:
        print(f"Error executing {e}, for {msg}")
        traceback.print_exc()
        ErrorResponse.query(str(e)).write(context.output)
        return context, Names.SYNC


def read_till_sync(context: Context):
    msg = context.mem.get('message', None)

    if not isinstance(msg, Sync):
        m, t = Message.read(context.input)

        if m.procesS_name not in {Names.SYNC, Names.CLOSE}:
            return context, Names.SYNC

    return context, Names.READY_FOR_QUERY


def write_extended_empty_response(context: Context):
    EmptyQueryResponse().write(context.output)

    return context, Names.RECEIVE_EXTENDED_QUERY_COMMAND


def write_extended_data_frame_response(context: Context):
    portal = context.mem['execution_portal']

    execution = portal.execution

    if not execution:
        raise Exception("We expect a execution_portal instance here")

    try:
        rows = DataFrameDataRows(df=execution.df,
                                 offset=execution.offset,
                                 max_rows=execution.max_rows).write(context.output)
        context.output.flush()

        df_rows = len(execution.df.index)

        read_all_rows = rows + execution.offset >= df_rows

        if rows == 0:
            EmptyQueryResponse().write(context.output)
        elif read_all_rows:
            CommandComplete(f"SELECT {df_rows}").write(context.output)
        elif rows < execution.max_rows:
            PortalSuspended().write(context.output)

        return context, Names.RECEIVE_EXTENDED_QUERY_COMMAND

    finally:
        context.output.flush()


def describe_statement_or_portal(context: Context):
    msg = context.mem.get('message', None)

    if not isinstance(msg, Describe):
        if isinstance(msg, Sync):
            return context, Names.SYNC

        raise Exception(f"Describe expected but got: {msg}")

    try:
        if msg.t == 'S':
            prepared_statement = context.prepared_statements[msg.name]
            print(f"Describe {prepared_statement}")
        else:
            portal = context.portals[msg.name]
            print(f"Describe {portal}")
            print(f"Writing DataFrameRowDescription")
            DataFrameRowDescription(df=portal.row_description).write(context.output)
            context.output.flush()

        return context, Names.RECEIVE_EXTENDED_QUERY_COMMAND
    except KeyError:
        print(
            f"No portal or prepared_statement_found for {msg.name} in {context.portals}, {context.prepared_statements}")
        return context, Names.ERROR
    except Exception as e:
        traceback.print_exc(e)
        print(f"describe_statement_or_portal {e}")
        ErrorResponse.query(str(e)).write(context.output)
        return context, Names.RECEIVE_EXTENDED_QUERY_COMMAND
